
--- Module implementing the LuaRocks "search" command.
-- Queries LuaRocks repositories.
module("luarocks.search", package.seeall)

local fs = require("luarocks.fs")
local path = require("luarocks.path")
local fetch = require("luarocks.fetch")
local deps = require("luarocks.deps")
local config = require("luarocks.config")

help_summary = "Query the Rocks repositories"

help = [[
<argument> is a substring of a rock name to search.
]]

--- Store a search result (a rock or rockspec) in the results table.
-- @param results table: The results table, where keys are package names and
-- versions are tables matching version strings to an array of repositories.
-- @param name string: Package name.
-- @param version string: Package version.
-- @param arch string: Architecture of rock ("all", "src" or platform
-- identifier), "rockspec" or "installed"
-- @param repo string: Pathname of a local repository of URL of
-- remote repository.
local function store_result(results, name, version, arch, repo)
   assert(type(results) == "table")
   assert(type(name) == "string")
   assert(type(version) == "string")
   assert(type(arch) == "string")
   assert(type(repo) == "string")
   
   if not results[name] then results[name] = {} end
   if not results[name][version] then results[name][version] = {} end
   table.insert(results[name][version], {
      arch = arch,
      repo = repo
   })
end

--- Test the name field of a query.
-- If query has a boolean field exact_name set to false,
-- then substring match is performed; otherwise, exact string
-- comparison is done.
-- @param query table: A query in dependency table format.
-- @param name string: A package name.
-- @return boolean: True if names match, false otherwise.
local function match_name(query, name)
   assert(type(query) == "table")
   assert(type(name) == "string")
   if query.exact_name == false then
      return name:match(query.name)
   else
      return name == query.name
   end
end

--- Store a match in a results table if version matches query.
-- Name, version, arch and repository path are stored in a given
-- table, optionally checking if version and arch (if given) match
-- a query.
-- @param results table: The results table, where keys are package names and
-- versions are tables matching version strings to an array of repositories.
-- @param repo string: URL or pathname of the repository.
-- @param name string: The name of the package being tested.
-- @param version string: The version of the package being tested.
-- @param arch string: The arch of the package being tested.
-- @param query table: A table describing the query in dependency
-- format (for example, {name = "filesystem", exact_name = false,
-- constraints = {op = "~>", version = {1,0}}}, arch = "rockspec").
-- If the arch field is omitted, the local architecture (config.arch)
-- is used. The special value "any" is also recognized, returning all
-- matches regardless of architecture.
local function store_if_match(results, repo, name, version, arch, query)
   if match_name(query, name) then
      if (not query.arch and (arch == "src" or arch == "all" or arch == "rockspec" or arch == "installed" or arch == config.arch))
      or query.arch == "any" or query.arch == arch
      then
         if deps.match_constraints(deps.parse_version(version), query.constraints) then
            store_result(results, name, version, arch, repo)
         end
      end
   end
end

--- Perform search on a local repository.
-- @param results table: The results table, where keys are package names and
-- versions are tables matching version strings to an array of repositories.
-- @param repo string: The pathname of the local repository.
-- @param query table: A table describing the query in dependency
-- format (for example, {name = "filesystem", exact_name = false,
-- constraints = {op = "~>", version = {1,0}}}, arch = "rockspec").
-- If the arch field is omitted, the local architecture (config.arch)
-- is used. The special value "any" is also recognized, returning all
-- matches regardless of architecture.
function disk_search(results, repo, query)
   assert(type(results) == "table")
   assert(type(repo) == "string")
   assert(type(query) == "table")
   
   for _, name in pairs(fs.dir(repo)) do
      local pathname = fs.make_path(repo, name)
      local rname, rversion, rarch = path.parse_rock_name(name)
      if not rname then
         rname, rversion, rarch = path.parse_rockspec_name(name)
      end
      if fs.is_dir(pathname) then
         for _, version in pairs(fs.dir(pathname)) do
            if version:match("-%d+$") then
               store_if_match(results, repo, name, version, "installed", query)
            end
         end
      elseif rname then
         store_if_match(results, repo, rname, rversion, rarch, query)
      end
   end
end

--- Perform search on a remote repository.
-- @param results table: The results table, where keys are package names and
-- versions are tables matching version strings to an array of repositories.
-- @param repo string: The URL of the remote repository.
-- @param query table: A table describing the query in dependency
-- format (for example, {name = "filesystem", exact_name = false,
-- constraints = {op = "~>", version = {1,0}}}, arch = "rockspec").
-- If the arch field is omitted, the local architecture (config.arch)
-- is used. The special value "any" is also recognized, returning all
-- matches regardless of architecture.
-- @return boolean or nil, string: true if no errors ocurred; or nil
-- followed by an error message.
function manifest_search(results, repo, query)
   assert(type(results) == "table")
   assert(type(repo) == "string")
   assert(type(query) == "table")
   
   local manifest, err = fetch.load_manifest(repo)
   if not manifest then
      return nil, err
   end
   for name, versions in pairs(manifest.repository) do
      for version, items in pairs(versions) do
         for _, item in ipairs(items) do
            store_if_match(results, repo, name, version, item.arch, query)
         end
      end
   end
   return true
end

--- Search on all configured repositories.
-- @param query table: A dependency query.
-- @return table or (nil, string): A table where keys are package names
-- and values are tables matching version strings to an array of
-- repositories; if no results are found, an empty table is returned.
-- In case of errors, nil and and error message are returned.
local function search_repos(query)
   assert(type(query) == "table")

   local results = {}
   for _, repo in ipairs(config.repositories) do
      local protocol, pathname = fs.split_url(repo)
      if protocol == "file" then
         repo = pathname
      end
      local ok, err = manifest_search(results, repo, query)
      if not ok then
         return nil, err
      end
   end
   return results
end

--- Prepare a query in dependency table format.
-- @param name string: The query name.
-- @param version string or nil: 
-- @return table: A query in table format
function make_query(name, version)
   assert(type(name) == "string")
   assert(type(version) == "string" or not version)
   
   local query = {
      name = name,
      constraints = {}
   }
   if version then
      table.insert(query.constraints, { op = "~>", version = deps.parse_version(version)})
   end
   return query
end

--- Get the URL for the latest in a set of versions.
-- @param name string: The package name to be used in the URL.
-- @param versions table: An array of version informations, as stored
-- in search results tables.
-- @return string or nil: the URL for the latest version if one could
-- be picked, or nil.
local function pick_latest_version(name, versions)
   assert(type(name) == "string")
   assert(type(versions) == "table")

   local vtables = {}
   for v, _ in pairs(versions) do
      table.insert(vtables, deps.parse_version(v))
   end
   table.sort(vtables)
   local version = vtables[#vtables].string
   local items = versions[version]
   if items then
      local pick = 1
      for i, item in ipairs(items) do
         if (item.arch == 'src' and items[pick].arch == 'rockspec')
         or (item.arch ~= 'src' and item.arch ~= 'rockspec') then
            pick = i
         end
      end
      return path.make_url(items[pick].repo, name, version, items[pick].arch)
   end
   return nil
end

--- Attempt to get a single URL for a given search.
-- @param query table: A dependency query.
-- @return string or table or (nil, string): URL for matching rock if
-- a single one was found, a table of candidates if it could not narrow to
-- a single result, or nil followed by an error message.
function find_suitable_rock(query)
   assert(type(query) == "table")
   
   local results, err = search_repos(query)
   if not results then
      return nil, err
   end
   local first = results and next(results)
   if first and next(results, first) == nil then
      return pick_latest_version(query.name, results[first])
   else
      return results
   end
end

--- Print a list of rocks/rockspecs on standard output.
-- @param results table: A table where keys are package names and versions
-- are tables matching version strings to an array of repositories.
-- @param show_repo boolean or nil: Whether to show repository
-- information or not. Default is true.
function print_results(results, show_repo)
   assert(type(results) == "table")
   assert(type(show_repo) == "boolean" or not show_repo)
   show_repo = show_repo or true

   for package, versions in pairs(results) do
      print()
      print(package)
      for version, repos in pairs(versions) do
         if show_repo then
            for _, repo in ipairs(repos) do
               print("   "..version.." ("..repo.arch..") - "..repo.repo)
            end
         else
            print("   "..version)
         end
      end
   end
   print()
end

--- Driver function for "search" command.
-- @param name string: A substring of a rock name to search.
-- @param version string or nil: a version may also be passed.
-- @return boolean or (nil, string): True if build was successful; nil and an
-- error message otherwise.
function run (name, version)
   if type(name) ~= "string" then
      return nil, "Argument missing, see help."
   end

   local query = make_query(name, version)
   query.exact_name = false
   local results = search_repos(query, false)
   print()
   print("Search results:")
   print("---------------")
   print_results(results, true)
   return true
end
